/* CPUID and RDRAND:

Copyright (c) 2014, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name of Intel Corporation nor the names of its
      contributors may be used to endorse or promote products derived from
      this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/


#include "os.h"
#include "intel_cpu.h"

#ifdef WINDOWS
    #include <limits.h>
    #include <intrin.h>
#endif // #ifdef WINDOWS


/** \brief Copyright (c) 2014, Intel Corporation: checks if this is an Intel CPU
 * -----------------------------------
 * returns 1 if the string "GenuineIntel" is found using the cpuid() function. Returns 0 otherwise.
 */
int is_intel_cpu ()
{
    static int intel_cpu= -1;
    cpuid_t info;

    /* So we don't call cpuid multiple times for the same information */

    if ( intel_cpu == -1 ) {
        cpuid(&info, 0, 0);

        if (
            memcmp((char *) &info.ebx, "Genu", 4) ||
            memcmp((char *) &info.edx, "ineI", 4) ||
            memcmp((char *) &info.ecx, "ntel", 4)
        ) {
            intel_cpu= 0;
        } else {
            intel_cpu= 1;
        }
    }

    return intel_cpu;
}

#ifdef __i386__
int have_cpuid ()
{
    /*
     * cpuid availability is determined by setting and clearing the
     * ID flag (bit 21) in the EFLAGS register. If we can do that, we
     * have cpuid. This is only necessary on 32-bit processors.
     */

    uint32_t fbefore, fafter;

    asm(" 					;\
                  pushf				;\
                  pushf				;\
                  pop %0				;\
                  mov %0,%1			;\
                  xor $0x40000,%1		;\
                  push %1				;\
                  popf				;\
                  pushf				;\
                  pop %1				;\
                  popf				"
        : "=&r" (fbefore), "=&r" (fafter)
       );

    return (0x40000 & (fbefore^fafter));
}
#endif

/** \brief Copyright (c) 2014, Intel Corporation: cpuid()
 * -----------------------------------
 * returns cpu info
 */
void cpuid (cpuid_t *info, unsigned int leaf, unsigned int subleaf)
{
#ifdef LINUX

#ifdef __i386__
    /* Can't use %ebx when compiling with -fPIC (or -fPIE) */

    asm volatile("	 	;\
                  mov %%ebx,%0 	;\
                  cpuid 			;\
                  xchgl %%ebx,%0	"
                 :	"=r" (info->ebx), "=a" (info->eax), "=c" (info->ecx), "=d" (info->edx)
                 :	"a" (leaf), "c" (subleaf)
                );
#else

    asm volatile("cpuid"
                 : "=a" (info->eax), "=b" (info->ebx), "=c" (info->ecx), "=d" (info->edx)
                 : "a" (leaf), "c" (subleaf)
                );
#endif

#endif  // #ifdef LINUX


#ifdef WINDOWS
    uint32_t regs[4];

    __cpuidex((int *)regs,(int)leaf,(int)subleaf);

    info->eax=regs[0];
    info->ebx=regs[1];
    info->ecx=regs[2];
    info->edx=regs[3];
#endif  // #ifdef WINDOWS
}



/** \brief check if a Digital Random Number Generator is available (Intel)
 * ---------------------------------------
 * Check if a Digital Random Number Generator is available (Intel)
 */
int get_drng_support()
{
    static int drng_features= -1;

    /* So we don't call cpuid multiple times for the same information */

    if ( drng_features == -1 ) {
        drng_features= DRNG_NO_SUPPORT;

#ifdef __i386__
        /* Older 32-bit processors may not even have cpuid (and they
         * will also not have a DRNG). */

        if ( ! have_cpuid() ) {
            return DRNG_NO_SUPPORT;
        }
#endif

        if ( is_intel_cpu() ) {
            cpuid_t info;

            cpuid(&info, 1, 0);

            if ( (info.ecx & 0x40000000) == 0x40000000 ) {
                drng_features|= DRNG_HAS_RDRAND;
            }

            cpuid(&info, 7, 0);

            if ( (info.ebx & 0x40000) == 0x40000 ) {
                drng_features|= DRNG_HAS_RDSEED;
            }
        }
    }

    return drng_features;
}

int rdrand32_step(uint32_t *rand)
{
#ifdef LINUX
    unsigned char ok;

    asm volatile ("rdrand %0; setc %1"
                  : "=r" (*rand), "=qm" (ok));

    return (int) ok;
#endif

#ifdef WINDOWS
    return _rdrand32_step(rand);
#endif  // #ifdef WINDOWS
}

int rdrand32_retry (unsigned int retries, uint32_t *rand)
{
	unsigned int count= 0;

	while ( count <= retries ) {
		if ( rdrand32_step(rand) ) {
			return 1;
		}

		++count;
	}

	return 0;
}


#ifdef __x86_64__
int rdrand64_step (uint64_t *rand)
{
    unsigned char ok;

    asm volatile ("rdrand %0; setc %1"
                  : "=r" (*rand), "=qm" (ok));

    return (int) ok;
}

int rdrand64_retry (unsigned int retries, uint64_t *rand)
{
	unsigned int count= 0;

	while ( count <= retries ) {
		if ( rdrand64_step(rand) ) {
			return 1;
		}

		++count;
	}

	return 0;
}
#endif
