
#include "os.h"

#ifdef WINDOWS
  #include <windows.h>
  #include <Wincrypt.h>                                     // for CryptGenRandom()
  #pragma comment(lib, "advapi32.lib")
#endif



#include <chrono>
#include <iostream>
#include "intel_cpu.h"
#include "rvb_math.h"
#include <stdio.h>



using namespace std::chrono;

namespace RVB
{

static int iDRNG_support = get_drng_support(); // bit 0 is set if RDRAND is supported
static FILE* fu = 0;


void close_file()
{
    fclose(fu);
}




/** \brief Random Source
 * ---------------------
 * /dev/random  and   /dev/urandom  are random sources
 * of the linux os claimed to be of high quality.
 *   - /dev/random  is blocking until material is available
 *   - /dev/urandom  is non blocking and thus preferable for our purposes
 */
void open_file()
{
    if(fu!=0) {
      return;
    }
    fu=fopen("/dev/urandom","r");
    if(!fu) {
      atexit(&close_file);
    }

}


/** \brief get a random value
 * -----------------------------
 * we generate truly random 64 bit values.

 * \return unsigned long long
*/
unsigned long long pre_random()
{
    unsigned long long ull_time,ull_drng;

    high_resolution_clock::time_point cl =high_resolution_clock::now();
    ull_time=cl.time_since_epoch().count(); // The number of ticks between timepoint cl and the clock's epoch

#ifdef LINUX
    // in 7 of 8 cases read from /dev/urandom
    if (ull_time & 0x0001000c) {
      open_file();
      fread(&ull_drng,1,sizeof(unsigned long long),fu);
    }
    else
      ull_drng=((unsigned long long)rand()+((unsigned long long)rand()<<25)+((unsigned long long)rand()<<35));
#endif // #ifdef LINUX
#ifdef WINDOWS
    HCRYPTPROV hProvider = 0;

    if (!::CryptAcquireContextW(&hProvider, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT))
      ull_drng=((unsigned long long)rand()+((unsigned long long)rand()<<27)+((unsigned long long)rand()<<35));

    const DWORD dwLength = (DWORD)sizeof(unsigned long long);
    if (!::CryptGenRandom(hProvider, dwLength, (BYTE *)&ull_drng))
    {
      ::CryptReleaseContext(hProvider, 0);
      ull_drng += ((unsigned long long)rand()+((unsigned long long)rand()<<26)+((unsigned long long)rand()<<35));
    }
    if (!::CryptReleaseContext(hProvider, 0))
      ull_drng ^= ((unsigned long long)rand()+((unsigned long long)rand()<<25)+((unsigned long long)rand()<<35));
#endif // #ifdef WINDOWS

    ull_time+=ull_drng;

    // support for Intel's hardware RNG (Ivy Bridge and later processor generations)
    if (iDRNG_support) {
#ifdef __x86_64__
        rdrand64_retry(10,(uint64_t*)&ull_drng);
#else
        unsigned int ui32;
        rdrand32_retry(10,&ui32);
        ull_drng=(unsigned long long)ui32;
        rdrand32_retry(10,&ui32);
        ull_drng += ((unsigned long long)ui32)<<32;
#endif
        ull_time += ull_drng;
    } // if (iDRNG_support)

    return ull_time;
}


} // namespace RVB







namespace RVB
{


static Modified_Kiss_2011 g_mod_kiss_2011;     // our static secondary pseudo random number generator (PRNG)


/** \brief Seeding an RNG with a huge Internal State
* ----------------------------------------
* for general use
*
* \param ullBias unsigned long long  64 bit bias value.
* \return void
*
*/
// Initialize the modified KISS 2011 PRNG with default seed and with content from dynamically allocated RAM
void Modified_Kiss_2011::randkiss_seed(unsigned long long ullBias)
{
  int i,j,k;
  kiss_j = QSIZE - 1;
  kiss_carry = 0;
  kiss_xs = 362436069362436069ULL;
  kiss_cng = 123456789987654321ULL;
  // Seed pQARY[] with CNG+XS+ullBias, but do NOT delete the initial state
  for (i = 0; i < QSIZE; i++) {
    // let's take advantage of /dev/urandom
    if (i<QSIZE/8) {
      ullBias+=pre_random();
      kiss_cng^=ullBias & 0x55555555aaaaaaaa;
    }
    else if ( (ullBias & 0x03)==0x00 )
      ullBias^=pre_random();
    else
      ullBias^=XS;

    QARY[i]^=CNG+XS+ullBias;

    // make the sequence unpredictable
    if ((ullBias & 0x00f1)==0x00) {
      ullBias^=CNG+pre_random();
      kiss_xs+=QARY[i];
    }
    if ((ullBias & 0x001f)==0x00) {
      ullBias^=XS-pre_random();
      kiss_cng-=QARY[i];
    }
  }

  // we allocate a multitude of small blocks. It is thus unlikely that
  // mmap is used to allocate page-aligned blocks that the kernel initializes to zero.
  // We then add the content of pQARY to the newly allocated RAM and free the blocks
  // ASAP
  size_t num_of_qwords_allocated=1024; // 8kB blocks
  unsigned long long * pUll[NUM_OF_SMALL_BLOCKS_TO_ALLOCATE];
  // 128*8192=1 MByte in total
  for (i=0; i<NUM_OF_SMALL_BLOCKS_TO_ALLOCATE; i++) {
    pUll[i]=(unsigned long long *)malloc(num_of_qwords_allocated*sizeof(unsigned long long));
  }
  k=0;
  for (i=0; i<NUM_OF_SMALL_BLOCKS_TO_ALLOCATE; i++) {
    if (pUll[i]) {
      // addresses can as well be used although there is not much randomness to expect
      ullBias+=(unsigned long long)pUll[i]^pre_random();
      // we simply add the content of the dynamically allocated memory to the bias and change pQARY[]
      for (j=0; j<(int)num_of_qwords_allocated; j++) {
        ullBias+=pUll[i][j];
        if ((ullBias & 0x07)==0x07) {
          ullBias^=XS;
        }
        else if ((ullBias & 0xf1)==0x00) {
          ullBias^=pre_random();
          QARY[k & (QSIZE-1)]-=ullBias;
        }
        else {
          QARY[k & (QSIZE-1)]+=ullBias;
          ullBias+=j; // in case this strategy fails, make sure that there is a slight movement
        }
        k++;
      }
    } // if (pUll[i])
    free(pUll[i]);
  } // for (i=0;i<NUM_OF_SMALL_BLOCKS_TO_ALLOCATE;i++)
}

Modified_Kiss_2011::Modified_Kiss_2011()
{
  QARY=(unsigned long long *)malloc(QSIZE*sizeof(unsigned long long));
  randkiss_seed((unsigned long long)QARY);
}
Modified_Kiss_2011::~Modified_Kiss_2011()
{
  if (QARY) {
    free(QARY);
  }
}

} // namespace RVB





namespace RVB
{

/** \brief Helper functions
* -------------------------
* Since most security options are defined in bits, but float precision has to be defined
* as decimal digits, we provide two helper functions to switch between the scales
*
* \param n size_t
* \return size_t
*
*/
size_t dec2bit(size_t n)
{
  return (size_t) ((double)n*3.32192809488+0.5);
}

size_t bit2dec(size_t n)
{
  return (size_t) ((double)n*0.301029995+0.5);
}



/** \brief String helper
* -------------------------
* Big number implementations are opaque to the user.
* Therefore the standard interfaces to the world are strings.
* These functions allow conversion.
*
* \param bi big_int const&
* \return string
*
*/
string itos(big_int const& bi)
{
  stringstream ss;
  ss << bi;
  return ss.str();
}


big_int stoi(string const& s)
{
  return big_int(s);
}


/** \brief Generating a 64 bit random integer
* ----------------------------------------
* for general use
*
* \param ullBias unsigned long long  Optional bias.
* \return unsigned long long  True random number (64 bits).
*
*/
unsigned long long get_random_value(unsigned long long ullBias=0LL)
{
  return g_mod_kiss_2011.kiss(pre_random()+rand()-ullBias);
}


/** \brief Generating a big random integer
* ----------------------------------------
* for usage as T polynomial index.
*
* \param bits size_t
* \return big_int
*
*/
big_int random_int(size_t bits)
{
  char const zif[] = "0123456789";
  bits=bit2dec(bits);
  unsigned char* buffer;
  unsigned long long ull=pre_random();
  buffer = new unsigned char[bits+10];
  for (int i=0;i<(int)bits+10;i++) {
    if (i & 0x01)
      buffer[i]=(char)(ull^g_mod_kiss_2011.kiss(ull+rand()));
    else
      buffer[i]=(char)(ull-g_mod_kiss_2011.kiss(ull-rand()));
    if (buffer[i] & 0x80)
      ull^=pre_random();
    else
      ull-=pre_random();
  }

  /** debug only */
  //    for(size_t j=0; j<bits; j++) {
  //        cout << std::hex << (unsigned)buffer[j] << " ";
  //    }

  string s;
  size_t i;
  for(i=0; i<bits; i++) {
    s=s+zif[buffer[i]%10];
  }
  while (s[0]=='0') {
    s[0]=zif[buffer[i++]%10];
    // let's better avoid possible but little probable crashes
    if ( (i>=bits+9) && (s[0]=='0') )
      s[0]='1';
  }
  delete buffer;
  return big_int(s);
}


} // namespace RVB

